﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Common;
using HeuristicLab.Problems.VehicleRouting.Interfaces;
using HeuristicLab.Problems.VehicleRouting.ProblemInstances;
using HeuristicLab.Problems.VehicleRouting.Variants;

namespace HeuristicLab.Problems.VehicleRouting.Encodings.Potvin {
  [Item("PushForwardInsertionCreator", "Creates a randomly initialized VRP solution.")]
  [StorableClass]
  public sealed class PushForwardInsertionCreator : PotvinCreator, IStochasticOperator {
    #region IStochasticOperator Members
    public ILookupParameter<IRandom> RandomParameter {
      get { return (LookupParameter<IRandom>)Parameters["Random"]; }
    }
    #endregion

    public IValueParameter<DoubleValue> Alpha {
      get { return (IValueParameter<DoubleValue>)Parameters["Alpha"]; }
    }
    public IValueParameter<DoubleValue> AlphaVariance {
      get { return (IValueParameter<DoubleValue>)Parameters["AlphaVariance"]; }
    }
    public IValueParameter<DoubleValue> Beta {
      get { return (IValueParameter<DoubleValue>)Parameters["Beta"]; }
    }
    public IValueParameter<DoubleValue> BetaVariance {
      get { return (IValueParameter<DoubleValue>)Parameters["BetaVariance"]; }
    }
    public IValueParameter<DoubleValue> Gamma {
      get { return (IValueParameter<DoubleValue>)Parameters["Gamma"]; }
    }
    public IValueParameter<DoubleValue> GammaVariance {
      get { return (IValueParameter<DoubleValue>)Parameters["GammaVariance"]; }
    }

    [StorableConstructor]
    private PushForwardInsertionCreator(bool deserializing) : base(deserializing) { }

    public PushForwardInsertionCreator()
      : base() {
      Parameters.Add(new LookupParameter<IRandom>("Random", "The pseudo random number generator."));
      Parameters.Add(new ValueParameter<DoubleValue>("Alpha", "The alpha value.", new DoubleValue(0.7)));
      Parameters.Add(new ValueParameter<DoubleValue>("AlphaVariance", "The alpha variance.", new DoubleValue(0.5)));
      Parameters.Add(new ValueParameter<DoubleValue>("Beta", "The beta value.", new DoubleValue(0.1)));
      Parameters.Add(new ValueParameter<DoubleValue>("BetaVariance", "The beta variance.", new DoubleValue(0.07)));
      Parameters.Add(new ValueParameter<DoubleValue>("Gamma", "The gamma value.", new DoubleValue(0.2)));
      Parameters.Add(new ValueParameter<DoubleValue>("GammaVariance", "The gamma variance.", new DoubleValue(0.14)));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new PushForwardInsertionCreator(this, cloner);
    }

    private PushForwardInsertionCreator(PushForwardInsertionCreator original, Cloner cloner)
      : base(original, cloner) {
    }

    // use the Box-Mueller transform in the polar form to generate a N(0,1) random variable out of two uniformly distributed random variables
    private static double Gauss(IRandom random) {
      double u = 0.0, v = 0.0, s = 0.0;
      do {
        u = (random.NextDouble() * 2) - 1;
        v = (random.NextDouble() * 2) - 1;
        s = Math.Sqrt(u * u + v * v);
      } while (s < Double.Epsilon || s > 1);
      return u * Math.Sqrt((-2.0 * Math.Log(s)) / s);
    }

    private static double N(double mu, double sigma, IRandom random) {
      return mu + (sigma * Gauss(random)); // transform the random variable sampled from N(0,1) to N(mu,sigma)
    }

    private static double GetDistance(int start, int end, IVRPProblemInstance problemInstance) {
      double distance = 0.0;

      double startX = problemInstance.Coordinates[start, 0];
      double startY = problemInstance.Coordinates[start, 1];

      double endX = problemInstance.Coordinates[end, 0];
      double endY = problemInstance.Coordinates[end, 1];

      distance =
          Math.Sqrt(
            Math.Pow(startX - endX, 2) +
            Math.Pow(startY - endY, 2));

      return distance;
    }

    private static PotvinEncoding CreateSolution(IVRPProblemInstance problemInstance, IRandom random,
      double alphaValue = 0.7, double betaValue = 0.1, double gammaValue = 0.2,
      double alphaVariance = 0.5, double betaVariance = 0.07, double gammaVariance = 0.14) {
      PotvinEncoding result = new PotvinEncoding(problemInstance);

      double alpha, beta, gamma;
      alpha = N(alphaValue, Math.Sqrt(alphaVariance), random);
      beta = N(betaValue, Math.Sqrt(betaVariance), random);
      gamma = N(gammaValue, Math.Sqrt(gammaVariance), random);

      int totalTours = 0;

      double distance = 0;
      double cost = 0;
      List<int> unroutedList = new List<int>();
      List<double> costList = new List<double>();

      IPickupAndDeliveryProblemInstance pdp = problemInstance as IPickupAndDeliveryProblemInstance;
      IMultiDepotProblemInstance mdp = problemInstance as IMultiDepotProblemInstance;

      int depots = 1;
      if (mdp != null) {
        depots = mdp.Depots.Value;
        for (int i = 0; i < result.VehicleAssignment.Length; i++)
          result.VehicleAssignment[i] = -1;
      }

      List<int> unrouted = new List<int>();
      for (int i = 1; i <= problemInstance.Cities.Value; i++)
        if (pdp == null || problemInstance.GetDemand(i) >= 0)
          unrouted.Add(i);

      for (int depot = 0; depot < depots; depot++) {
        int vehicleCount = problemInstance.Vehicles.Value;
        List<int> vehicles = new List<int>();
        if (mdp != null) {
          vehicleCount = 0;
          for (int i = 0; i < mdp.VehicleDepotAssignment.Length; i++) {
            if (mdp.VehicleDepotAssignment[i] == depot) {
              vehicleCount++;
              vehicles.Add(i);
            }
          }
        } else {
          for (int i = 0; i < problemInstance.Vehicles.Value; i++)
            vehicles.Add(i);
        }

        costList.Clear();
        unroutedList.Clear();

        double x0 = problemInstance.Coordinates[depot, 0];
        double y0 = problemInstance.Coordinates[depot, 1];

        /*-----------------------------------------------------------------------------
        * generate cost list
        *-----------------------------------------------------------------------------
        */
        for (int i = 1; i <= problemInstance.Cities.Value; i++) {
          //only insert sources
          if (unrouted.Contains(i) && (pdp == null || problemInstance.GetDemand(i) >= 0)) {
            distance = GetDistance(i + depots - 1, depot, problemInstance);
            if (problemInstance.Coordinates[i + depots - 1, 0] < x0) distance = -distance;

            double dueTime = 0;
            if (problemInstance is ITimeWindowedProblemInstance)
              dueTime = (problemInstance as ITimeWindowedProblemInstance).DueTime[i + depots - 1];

            cost = -alpha * distance + // distance 0 <-> City[i]
                   beta * dueTime + // latest arrival time
                   gamma * (Math.Asin((problemInstance.Coordinates[i + depots - 1, 0] - y0) / distance) / 360 * distance); // polar angle

            int index = 0;
            while (index < costList.Count && costList[index] < cost) index++;

            costList.Insert(index, cost);

            unroutedList.Insert(index, i);
          }
        }

        int tours = 0;
        double minimumCost = double.MaxValue;
        int indexOfMinimumCost = -1;
        int indexOfMinimumCost2 = -1;
        int indexOfCustomer = -1;

        Tour tour = new Tour();
        result.Tours.Add(tour);
        result.VehicleAssignment[totalTours] = vehicles[0];
        vehicles.RemoveAt(0);
        tours++;
        totalTours++;

        tour.Stops.Add(unroutedList[0]);

        if (pdp != null) {
          tour.Stops.Add(pdp.GetPickupDeliveryLocation(unroutedList[0]));
        }

        unrouted.Remove(unroutedList[0]);
        unroutedList.RemoveAt(0);

        while (unroutedList.Count > 0 && (tours < vehicleCount || depot == depots - 1)) {
          for (int c = 0; c < unroutedList.Count; c++) {
            VRPEvaluation eval = problemInstance.EvaluateTour(tour, result);
            double originalCosts = eval.Quality;
            int customer = unroutedList[c];

            for (int i = 0; i <= tour.Stops.Count; i++) {
              tour.Stops.Insert(i, customer);
              eval = problemInstance.EvaluateTour(tour, result);
              double tourCost = eval.Quality - originalCosts;

              if (pdp != null) {
                for (int j = i + 1; j <= tour.Stops.Count; j++) {
                  bool feasible;
                  cost = tourCost +
                    problemInstance.GetInsertionCosts(eval, result, pdp.GetPickupDeliveryLocation(unroutedList[c]), 0, j, out feasible);
                  if (cost < minimumCost && feasible) {
                    minimumCost = cost;
                    indexOfMinimumCost = i;
                    indexOfMinimumCost2 = j;
                    indexOfCustomer = c;
                  }
                }
              } else {
                cost = tourCost;
                bool feasible = problemInstance.Feasible(eval);
                if (cost < minimumCost && feasible) {
                  minimumCost = cost;
                  indexOfMinimumCost = i;
                  indexOfCustomer = c;
                }
              }

              tour.Stops.RemoveAt(i);
            }
          }

          // insert customer if found
          if (indexOfMinimumCost != -1) {
            tour.Stops.Insert(indexOfMinimumCost, unroutedList[indexOfCustomer]);
            if (pdp != null) {
              tour.Stops.Insert(indexOfMinimumCost2, pdp.GetPickupDeliveryLocation(unroutedList[indexOfCustomer]));
            }

            unrouted.Remove(unroutedList[indexOfCustomer]);
            unroutedList.RemoveAt(indexOfCustomer);
          } else if (tours < vehicleCount || depot == depots - 1) { // no feasible customer found
            tour = new Tour();
            result.Tours.Add(tour);
            result.VehicleAssignment[totalTours] = vehicles[0];
            vehicles.RemoveAt(0);
            tours++;
            totalTours++;

            tour.Stops.Add(unroutedList[0]);
            if (pdp != null) {
              tour.Stops.Add(pdp.GetPickupDeliveryLocation(unroutedList[0]));
            }

            unrouted.Remove(unroutedList[0]);
            unroutedList.RemoveAt(0);
          }
          // reset minimum		
          minimumCost = double.MaxValue;
          indexOfMinimumCost = -1;
          indexOfMinimumCost2 = -1;
          indexOfCustomer = -1;
        }
      }

      if (mdp != null) {
        List<int> availableVehicles = new List<int>();
        for (int i = 0; i < mdp.Vehicles.Value; i++)
          availableVehicles.Add(i);

        for (int i = 0; i < result.VehicleAssignment.Length; i++) {
          if (result.VehicleAssignment[i] != -1)
            availableVehicles.Remove(result.VehicleAssignment[i]);
        }

        for (int i = 0; i < result.VehicleAssignment.Length; i++) {
          if (result.VehicleAssignment[i] == -1) {
            result.VehicleAssignment[i] = availableVehicles[0];
            availableVehicles.RemoveAt(0);
          }
        }
      }

      return result;
    }

    public override IOperation Apply() {
      VRPToursParameter.ActualValue = CreateSolution(ProblemInstance, RandomParameter.ActualValue,
        Alpha.Value.Value, Beta.Value.Value, Gamma.Value.Value,
        AlphaVariance.Value.Value, BetaVariance.Value.Value, GammaVariance.Value.Value);

      return base.Apply();
    }
  }
}
